#Code to reproduce results of manuscript "A New Dimension to Mortality Forecasts" by Benjamin Seligman,
#Gabi Greenberg, and Shripad Tuljapurkar, Demographic Research, 2016

#First vs. Second-Stage Forecasts, different base periods, Females

###Trends for Females
#Fitting LC Model 1950-1990, then 1960-1990, and forecasting e0
#Second-Stage Estimation of K
#Objective function to optimize:
stage2<-function(k,a,b,N,D){
  Diff = abs(D-sum(exp(a+b*k)*N))
}
#Producing kt for G7 Countries using the HMD
namevec<-c("CAN","DEUTW","FRATNP","GBR_NP","ITA","JPN","USA")
for(i in 1:length(namevec)){
  #converts full life-tables into matrix of mortality rates
  #with rows being age groups and columns being the year of the data
  stuff<-read.table(paste(namevec[i],".fltper_1x1.txt",sep=""),header=TRUE,skip=2)
  if(stuff$Year[1] < 1950){
    stuff = stuff[which(stuff$Year > 1949),]
  }
  #Matrix of life expectancies
  assign(paste(namevec[i],".e0",sep=""),cbind(as.numeric(levels(factor(stuff$Year))),as.numeric(stuff$ex[which(stuff$Age==0)])))
  #Post-Launch actual log mortalities
  assign(paste(namevec[i],".logm.post",sep=""),log(stuff$mx[which(stuff$Year>1989)]))
  stuff<-stuff[which(stuff$Year<1991),]
  W<-length(levels(factor(stuff$Age)))
  #Initial log mortality rates
  assign(paste(namevec[i],".logm",sep=""),log(stuff$mx[which(stuff$Year==1990)]))
  #Calculating the Lee-Carter Model for 1950-1990
  #Converting matrix of serial lifetables into matrix of mortality rates by age (rows) & period/year (columns)
  Q<-vector(mode="numeric")
  Years<-as.integer(levels(factor(stuff$Year)))
  for(z in 1:length(Years)){
    Q<-cbind(Q,stuff[(z*W-(W-1)):(z*W),3])
  }
  #LC Model Calculations
  logQ<-log(Q)
  A<-rowMeans(logQ)
  X<-logQ-A
  S<-svd(X)
  Vt<-t(S$v)
  Ut<-t(S$u)
  B<-Ut[1,]/sum(Ut[1,])
  #B<-Vt[,1]/sum(Vt[,1]) <-This uses the column space (years), which is not correct
  k<-S$v[,1]*sum(Ut[1,])*S$d[1]
  assign(paste(namevec[i],".50",sep=""),list(a=A,b=B,k=cbind(Years,k)))
  #2nd stage k's for 1950-1990
  Pop<-read.table(paste(namevec[i],".Population.txt",sep=""),header=T,skip=2)
  Pop=Pop[(substr(Pop$Year,5,5)!="+"),]
  Pop$Year=as.numeric(substr(Pop$Year,1,4))
  Pop<-Pop[which(Pop$Year>1949),]
  Pop<-Pop[which(Pop$Year<1991),]
  out<-vector(mode="numeric",length=length(k))
  rang<-c(2*min(k),2*max(k))
  W<-length(A)
  for(y in 1:length(k)){
    N<-Pop[(W*y-(W-1)):(W*y),3]
    D<-Pop[(W*y-(W-1)):(W*y),3]%*%stuff[(W*y-(W-1)):(W*y),3]
    out[y]<-as.numeric(optimize(f = stage2,rang,a=A,b=B,N=N,D=D)[1])
  }
  assign(paste(namevec[i],".2ndk50",sep=""),out)
  ##################################################
  #Recalculating the Lee-Carter Model for 1960-1990
  #Converting matrix of serial lifetables into matrix of mortality rates by age (rows) & period/year (columns)
  stuff<-stuff[which(stuff$Year>1959),]
  Q<-vector(mode="numeric")
  Years<-as.integer(levels(factor(stuff$Year)))
  for(z in 1:length(Years)){
    Q<-cbind(Q,stuff[(z*W-(W-1)):(z*W),3])
  }
  #LC Model Calculations
  logQ<-log(Q)
  A<-rowMeans(logQ)
  X<-logQ-A
  S<-svd(X)
  Vt<-t(S$v)
  Ut<-t(S$u)
  B<-Ut[1,]/sum(Ut[1,])
  #B<-Vt[,1]/sum(Vt[,1]) <-This uses the column space (years), which is not correct
  k<-S$v[,1]*sum(Ut[1,])*S$d[1]
  assign(paste(namevec[i],".60",sep=""),list(a=A,b=B,k=cbind(Years,k)))
  #2nd stage k's for 1960-1990
  Pop<-Pop[which(Pop$Year>1959),]
  out<-vector(mode="numeric",length=length(k))
  rang<-c(2*min(k),2*max(k))
  W<-length(A)
  for(y in 1:length(k)){
    N<-Pop[(W*y-(W-1)):(W*y),3]
    D<-Pop[(W*y-(W-1)):(W*y),3]%*%stuff[(W*y-(W-1)):(W*y),3]
    out[y]<-as.numeric(optimize(f = stage2,rang,a=A,b=B,N=N,D=D)[1])
  }
  assign(paste(namevec[i],".2ndk60",sep=""),out)
}


#forecasting kt forward 20 years
library(forecast)
for(i in 1:length(namevec)){
  assign(paste(namevec[i],".fc50",sep=""),rwf(get(paste(namevec[i],".50",sep=""))$k[,2],h=20,drift=T))
  assign(paste(namevec[i],".fc50.2ndk",sep=""),rwf(get(paste(namevec[i],".2ndk50",sep="")),h=20,drift=T))
  assign(paste(namevec[i],".fc60",sep=""),rwf(get(paste(namevec[i],".60",sep=""))$k[,2],h=20,drift=T))
  assign(paste(namevec[i],".fc60.2ndk",sep=""),rwf(get(paste(namevec[i],".2ndk60",sep="")),h=20,drift=T))
}

#Calculating e0 using rwf forecasts without relying recomposition with SVD
#function to project forward mx given initial mx, series of k's, and b
project<-function(init,b,k){
  delts<-tail(k,-1)-head(k,-1)
  q<-as.matrix(init+b*delts[1])
  for(i in 2:length(delts)){q<-cbind(q,q[,(i-1)]+b*delts[i])}
  return(q)
}

#function to calculate life expectancies from mx
e0calc<-function(m1){
  m<-exp(as.matrix(m1))
  q1<-m[1,]/(1+m[1,]*(1-(0.053+2.8*m[1,])))
  q<-m[2:nrow(m),]/(1+m[2:nrow(m),]*0.5)
  q<-c(q1,q)
  lx<-as.matrix(1-q)
  l=1
  for(p in 2:length(q)){
    l<-rbind(l,apply(head(lx,(p-1)),2,prod))
  }
  e0<-sum((1-q)*l+q*l*c(0.07 + 1.7*m[1],rep(0.5,110)))
  return(e0)
}

#Converting actual logms for 1990-most recent year to e0's
for(i in 1:length(namevec)){
  logmi<-get(paste(namevec[i],".logm.post",sep=""))
  logm<-vector(mode="numeric")
  for(k in 1:(length(logmi)/111)){logm<-cbind(logm,logmi[(k*111-110):(k*111)])}
  e0<-vector(mode="numeric")
  for(k in 1:ncol(logm)){
    e0<-c(e0,e0calc(logm[,k]))
  }
  assign(paste(namevec[i],".logm.e0",sep=""),e0)
}

#Projecting forward and calculating appropriate e0's w/ 1st stage k's
for(i in 1:length(namevec)){
  #Appropriate b vector
  b50<-get(paste(namevec[i],".50",sep=""))$b
  #extracts k's for 1990 from data for launch point
  k<-tail(get(paste(namevec[i],".50",sep=""))$k[,2],1)
  #Various k's
  k50<-list(mean=c(k,get(paste(namevec[i],".fc50",sep=""))$mean),lower=c(k,get(paste(namevec[i],".fc50",sep=""))$lower[,2]),upper=c(k,get(paste(namevec[i],".fc50",sep=""))$upper[,2]))
  #Contains elements of the projection made
  proj50<-as.matrix(get(paste(namevec[i],".logm",sep="")))
  proj50<-list(mean=proj50,lower=proj50,upper=proj50)
  #contains life expectancy calculations
  expect50<-list("mean","lower","upper")
  #Repeat of above
  b60<-get(paste(namevec[i],".60",sep=""))$b
  k<-tail(get(paste(namevec[i],".60",sep=""))$k[,2],1)
  k60<-list(mean=c(k,get(paste(namevec[i],".fc60",sep=""))$mean),lower=c(k,get(paste(namevec[i],".fc60",sep=""))$lower[,2]),upper=c(k,get(paste(namevec[i],".fc60",sep=""))$upper[,2]))
  proj60<-as.matrix(get(paste(namevec[i],".logm",sep="")))
  proj60<-list(mean=proj60,lower=proj60,upper=proj60)
  expect60<-list("mean","lower","upper")
  #running the projection itself --> get vectors of 
  for(j in 1:3){
    proj50[[j]]<-cbind(proj50[[j]],project(proj50[[j]],b50,k50[[j]]))
    for(h in 1:21){expect50[[j]][h]<-e0calc(proj50[[j]][,h])}
    proj60[[j]]<-cbind(proj60[[j]],project(proj60[[j]],b60,k60[[j]]))
    for(h in 1:21){expect60[[j]][h]<-e0calc(proj60[[j]][,h])}
  }
  assign(paste(namevec[i],".e0.50",sep=""),expect50)
  plot(x=seq(1990,2010,1),y=expect50[[1]],type="l",col="red",ylim=as.numeric(c(min(expect50[[1]],expect50[[2]],expect50[[3]]),max(expect50[[1]],expect50[[2]],expect50[[3]]))),xlab="Year",ylab="e0",main=paste(namevec[i],"Projected and Actual e0, 1950 base, 1st stage k, Females"))
  lines(x=seq(1990,2010,1),y=expect50[[2]],col="red")
  lines(x=seq(1990,2010,1),y=expect50[[3]],col="red")
  lines(x=tail(get(paste(namevec[i],".e0",sep=""))[,1],20),y=tail(get(paste(namevec[i],".e0",sep=""))[,2],20),col="blue")
  lines(x=seq(1990,(1989+length(get(paste(namevec[i],".logm.e0",sep="")))),1),y=get(paste(namevec[i],".logm.e0",sep="")),col="green")
  assign(paste(namevec[i],".e0.60",sep=""),expect60)
  plot(x=seq(1990,2010,1),y=expect60[[1]],type="l",col="red",ylim=as.numeric(c(min(expect60[[1]],expect60[[2]],expect60[[3]]),max(expect60[[1]],expect60[[2]],expect60[[3]]))),xlab="Year",ylab="e0",main=paste(namevec[i],"Projected and Actual e0,1960 base, 1st stage k, Females"))
  lines(x=seq(1990,2010,1),y=expect60[[2]],col="red")
  lines(x=seq(1990,2010,1),y=expect60[[3]],col="red")
  lines(x=tail(get(paste(namevec[i],".e0",sep=""))[,1],20),y=tail(get(paste(namevec[i],".e0",sep=""))[,2],20),col="blue")
  lines(x=seq(1990,(1989+length(get(paste(namevec[i],".logm.e0",sep="")))),1),y=get(paste(namevec[i],".logm.e0",sep="")),col="green")
}

#Repeating above w/ 2nd-stage k's
for(i in 1:length(namevec)){
  #Appropriate b vector
  b50<-get(paste(namevec[i],".50",sep=""))$b
  #extracts k's for 1990 from data for launch point
  k<-tail(get(paste(namevec[i],".2ndk50",sep="")),1)
  #Various k's
  k50<-list(mean=c(k,get(paste(namevec[i],".fc50.2ndk",sep=""))$mean),lower=c(k,get(paste(namevec[i],".fc50.2ndk",sep=""))$lower[,2]),upper=c(k,get(paste(namevec[i],".fc50.2ndk",sep=""))$upper[,2]))
  #Contains elements of the projection made
  proj50<-as.matrix(get(paste(namevec[i],".logm",sep="")))
  proj50<-list(mean=proj50,lower=proj50,upper=proj50)
  #contains life expectancy calculations
  expect50<-list("mean","lower","upper")
  #Repeat of above
  b60<-get(paste(namevec[i],".60",sep=""))$b
  k<-tail(get(paste(namevec[i],".2ndk60",sep="")),1)
  k60<-list(mean=c(k,get(paste(namevec[i],".fc60.2ndk",sep=""))$mean),lower=c(k,get(paste(namevec[i],".fc60.2ndk",sep=""))$lower[,2]),upper=c(k,get(paste(namevec[i],".fc60.2ndk",sep=""))$upper[,2]))
  proj60<-as.matrix(get(paste(namevec[i],".logm",sep="")))
  proj60<-list(mean=proj60,lower=proj60,upper=proj60)
  expect60<-list("mean","lower","upper")
  #running the projection itself --> get vectors of 
  for(j in 1:3){
    proj50[[j]]<-cbind(proj50[[j]],project(proj50[[j]],b50,k50[[j]]))
    for(h in 1:21){expect50[[j]][h]<-e0calc(proj50[[j]][,h])}
    proj60[[j]]<-cbind(proj60[[j]],project(proj60[[j]],b60,k60[[j]]))
    for(h in 1:21){expect60[[j]][h]<-e0calc(proj60[[j]][,h])}
  }
  assign(paste(namevec[i],".e0.50.2ndk",sep=""),expect50)
  plot(x=seq(1990,2010,1),y=expect50[[1]],type="l",col="red",ylim=as.numeric(c(min(expect50[[1]],expect50[[2]],expect50[[3]]),max(expect50[[1]],expect50[[2]],expect50[[3]]))),xlab="Year",ylab="e0",main=paste(namevec[i],"Projected and Actual e0, 1950 base, 2nd stage k, Females"))
  lines(x=seq(1990,2010,1),y=expect50[[2]],col="red")
  lines(x=seq(1990,2010,1),y=expect50[[3]],col="red")
  lines(x=tail(get(paste(namevec[i],".e0",sep=""))[,1],20),y=tail(get(paste(namevec[i],".e0",sep=""))[,2],20),col="blue")
  lines(x=seq(1990,(1989+length(get(paste(namevec[i],".logm.e0",sep="")))),1),y=get(paste(namevec[i],".logm.e0",sep="")),col="green")
  assign(paste(namevec[i],".e0.60.2ndk",sep=""),expect60)
  plot(x=seq(1990,2010,1),y=expect60[[1]],type="l",col="red",ylim=as.numeric(c(min(expect60[[1]],expect60[[2]],expect60[[3]]),max(expect60[[1]],expect60[[2]],expect60[[3]]))),xlab="Year",ylab="e0",main=paste(namevec[i],"Projected and Actual e0,1960 base, 2nd stage k, Females"))
  lines(x=seq(1990,2010,1),y=expect60[[2]],col="red")
  lines(x=seq(1990,2010,1),y=expect60[[3]],col="red")
  lines(x=tail(get(paste(namevec[i],".e0",sep=""))[,1],20),y=tail(get(paste(namevec[i],".e0",sep=""))[,2],20),col="blue")
  lines(x=seq(1990,(1989+length(get(paste(namevec[i],".logm.e0",sep="")))),1),y=get(paste(namevec[i],".logm.e0",sep="")),col="green")
}
